// Project : K-means
// Author : Dario GENGA

#ifndef _KMEANS_M
#define _KMEANS_M

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>

/// A group who contains points.
typedef struct _cluster {
    /// The center of gravity of the cluster.
    struct _point* centroid;
    /// The color representing the cluster.
    int color;
} cluster;

/// A point in the universe, which represents a data.
typedef struct _point {
    /// The coordinates of the point.
    double* value;
    /// The cluster who contains the point.
    struct _cluster* cluster;
    /// The color to use to draw the point.
    int color;
    /// A label representing the group to which belongs te point. Only used for testing.
    char* label;
} point;

/// A universe who contains data, of points, and clusters who contains the data.
typedef struct _kmeans {
    /// The number of clusters in the universe.
    int k;
    /// A 'k' size array of clusters, who contains the data.
    struct _cluster* clusters_array;
    /// An array of points, representing the data in the universe.
    struct _point** points_array;
    /// The number of points in the universe.
    int nb_points;
} kmeans;

/// Create the kmeans universe.
/// \param k The number of clusters in the universe.
/// \param data The array of points
/// \param nb_points The number of points in the universe.
/// \return The kmeans object of the universe.
kmeans kmeans_create(int k, point** data, int nb_points);

/// Create a point.
/// \param value The coordinates of the point.
/// \return The point objet initialized with its coordinates
point point_create(double* value);

/// Create a cluster.
/// \param centroid The point representing the center of gravity of the cluster.
/// \return The cluster object.
kmeans cluster_create(point centroid);

/// Initialize the universe with the file specified in argument when starting the program. If no arguments has been
/// specified, use the standard input instead.
/// \param universe The universe to initialize.
void init_from_cmd_arguments(kmeans *universe);

/// Load the universe with the data in the file source.
/// \param universe The universe that will contains the data.
/// \param source_file The path to the file to be read.
void read_data_source(kmeans *universe, char* source_file);

/// Save the universe and its data in the file specified.
/// \param universe The universe to save.
/// \param output_file The path of the file where we save the universe. Create the file if it doesn't exist, else write over it.
void write_data_output(kmeans *universe, char* output_file);

/// Initialize the clusters in the universe by determining their center of gravity.
/// \param universe The universe who contains the clusters.
void init_clusters(kmeans *universe);

/// Compute the distance between two points of same dimensions.
/// \param p1 The first point.
/// \param p2 The second point.
/// \return The distance between the two points.
double compute_distance(point* p1, point p2);

/// Calculate the position of the center of gravity of the cluster.
/// \param clstr The cluster.
/// \param universe The universe who contains the points of the cluster.
void compute_center_of_gravity(cluster* clstr, kmeans* universe);

/// Assign the point to the most coherent cluster.
/// \param p The point to be assigned.
/// \param universe The universe who contains the point and the clusters.
void assign_points_to_cluster(point* p, kmeans* universe);

/// Start the clustering with the kmeans algorithm.
/// \param universe The kmeans universe.
void start_clustering(kmeans* universe);

/// Free the memory of the cluster.
/// \param clstr The cluster to free.
void destroy_cluster(cluster* clstr);

/// Free the memory of the point.
/// \param p The point to free.
void destroy_point(point* p);

/// Free the memory of universe.
/// \param kmeans The universe to free.
void destroy_universe(kmeans kmeans);

#endif
